# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.enums import (
    Method,
    Mode,
    Test,
    Weight,
    label_to_method,
    label_to_mode,
    label_to_test,
    label_to_weight,
    method_to_label,
    mode_to_label,
    test_to_label,
    weight_to_label,
)
from deareis.data import (
    FitSettings,
    SimulationSettings,
    TestSettings,
)
from deareis.config import (
    DEFAULT_TEST_SETTINGS,
    DEFAULT_FIT_SETTINGS,
    DEFAULT_SIMULATION_SETTINGS,
)
from deareis.gui.plotting.export import (
    EXTENSIONS,
    LEGEND_LOCATIONS,
    PREVIEW_LIMITS,
    UNITS_PER_INCH,
)
from deareis.signals import Signal
import deareis.signals as signals


def section_spacer():
    dpg.add_spacer(height=8)


def general_settings(label_pad: int, state):
    with dpg.collapsing_header(label="General", default_open=True):
        auto_backup_interval: int = dpg.generate_uuid()

        def update_auto_backup_interval(value: int):
            state.config.auto_backup_interval = value

        with dpg.group(horizontal=True):
            dpg.add_text("Auto-backup interval".rjust(label_pad))
            attach_tooltip(tooltips.general.auto_backup_interval)
            dpg.add_input_int(
                default_value=state.config.auto_backup_interval,
                label="actions",
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=lambda s, a, u: update_auto_backup_interval(a),
                width=-54,
                tag=auto_backup_interval,
            )
        section_spacer()


def kramers_kronig_tab_settings(label_pad: int, state):
    with dpg.collapsing_header(label="Kramers-Kronig tab", default_open=True):
        test_test_combo: int = dpg.generate_uuid()
        test_mode_combo: int = dpg.generate_uuid()
        test_mu_crit_slider: int = dpg.generate_uuid()
        test_add_cap_checkbox: int = dpg.generate_uuid()
        test_add_ind_checkbox: int = dpg.generate_uuid()
        test_method_combo: int = dpg.generate_uuid()
        test_max_nfev_input: int = dpg.generate_uuid()

        def update_default_test_settings():
            state.config.default_test_settings = TestSettings(
                label_to_test.get(dpg.get_value(test_test_combo), Test.COMPLEX),
                label_to_mode.get(dpg.get_value(test_mode_combo), Mode.EXPLORATORY),
                999,
                dpg.get_value(test_mu_crit_slider),
                dpg.get_value(test_add_cap_checkbox),
                dpg.get_value(test_add_ind_checkbox),
                label_to_method.get(dpg.get_value(test_method_combo), Method.LEASTSQ),
                dpg.get_value(test_max_nfev_input),
            )

        def restore_default_test_settings():
            dpg.set_value(test_test_combo, test_to_label[DEFAULT_TEST_SETTINGS.test])
            dpg.set_value(test_mode_combo, mode_to_label[DEFAULT_TEST_SETTINGS.mode])
            dpg.set_value(test_mu_crit_slider, DEFAULT_TEST_SETTINGS.mu_criterion)
            dpg.set_value(test_add_cap_checkbox, DEFAULT_TEST_SETTINGS.add_capacitance)
            dpg.set_value(test_add_ind_checkbox, DEFAULT_TEST_SETTINGS.add_inductance)
            dpg.set_value(
                test_method_combo, method_to_label[DEFAULT_TEST_SETTINGS.method]
            )
            dpg.set_value(test_max_nfev_input, DEFAULT_TEST_SETTINGS.max_nfev)
            update_default_test_settings()

        with dpg.group(horizontal=True):
            dpg.add_text("Test".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.test)
            dpg.add_combo(
                items=list(label_to_test.keys()),
                default_value=test_to_label.get(
                    state.config.default_test_settings.test
                ),
                callback=update_default_test_settings,
                tag=test_test_combo,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Mode".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.mode)
            dpg.add_combo(
                items=list(label_to_mode.keys()),
                default_value=mode_to_label.get(
                    state.config.default_test_settings.mode
                ),
                callback=update_default_test_settings,
                tag=test_mode_combo,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Add capacitor".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.add_capacitance)
            dpg.add_checkbox(
                default_value=state.config.default_test_settings.add_capacitance,
                callback=update_default_test_settings,
                tag=test_add_cap_checkbox,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Add inductor".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.add_inductance)
            dpg.add_checkbox(
                default_value=state.config.default_test_settings.add_inductance,
                callback=update_default_test_settings,
                tag=test_add_ind_checkbox,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Mu-criterion".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.mu_criterion)
            dpg.add_slider_float(
                default_value=state.config.default_test_settings.mu_criterion,
                min_value=0.01,
                max_value=0.99,
                clamped=True,
                format="%.2f",
                callback=update_default_test_settings,
                tag=test_mu_crit_slider,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Fitting method".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.method)
            dpg.add_combo(
                items=list(label_to_method.keys()),
                default_value=method_to_label.get(
                    state.config.default_test_settings.method
                ),
                callback=update_default_test_settings,
                tag=test_method_combo,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Max. num. func. eval.".rjust(label_pad))
            attach_tooltip(tooltips.kramers_kronig.nfev)
            dpg.add_input_int(
                default_value=state.config.default_test_settings.max_nfev,
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=update_default_test_settings,
                tag=test_max_nfev_input,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=restore_default_test_settings,
            )
        section_spacer()


def fitting_tab_settings(label_pad: int, state):
    with dpg.collapsing_header(label="Fitting tab", default_open=True):
        fit_cdc_input: int = dpg.generate_uuid()
        fit_method_combo: int = dpg.generate_uuid()
        fit_weight_combo: int = dpg.generate_uuid()
        fit_max_nfev_input: int = dpg.generate_uuid()

        def update_default_fit_settings():
            state.config.default_fit_settings = FitSettings(
                dpg.get_value(fit_cdc_input),
                label_to_method.get(dpg.get_value(fit_method_combo), Method.AUTO),
                label_to_weight.get(dpg.get_value(fit_weight_combo), Weight.AUTO),
                dpg.get_value(fit_max_nfev_input),
            )

        def restore_default_fit_settings():
            dpg.set_value(fit_cdc_input, DEFAULT_FIT_SETTINGS.cdc)
            dpg.set_value(
                fit_method_combo,
                method_to_label.get(DEFAULT_FIT_SETTINGS.method, "Auto"),
            )
            dpg.set_value(
                fit_weight_combo,
                weight_to_label.get(DEFAULT_FIT_SETTINGS.weight, "Auto"),
            )
            dpg.set_value(fit_max_nfev_input, DEFAULT_FIT_SETTINGS.max_nfev)
            update_default_fit_settings()

        with dpg.group(horizontal=True):
            dpg.add_text("Circuit".rjust(label_pad))
            attach_tooltip(tooltips.fitting.cdc)
            dpg.add_input_text(
                default_value=state.config.default_fit_settings.cdc,
                on_enter=True,
                callback=update_default_fit_settings,
                width=-1,
                tag=fit_cdc_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Method".rjust(label_pad))
            attach_tooltip(tooltips.fitting.method)
            dpg.add_combo(
                items=["Auto"] + list(label_to_method.keys()),
                default_value=method_to_label.get(
                    state.config.default_fit_settings.method, "Auto"
                ),
                width=-1,
                callback=update_default_fit_settings,
                tag=fit_method_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Weight function".rjust(label_pad))
            attach_tooltip(tooltips.fitting.weight)
            dpg.add_combo(
                items=["Auto"] + list(label_to_weight.keys()),
                default_value=weight_to_label.get(
                    state.config.default_fit_settings.weight, "Auto"
                ),
                width=-1,
                callback=update_default_fit_settings,
                tag=fit_weight_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Max. num. func. eval.".rjust(label_pad))
            attach_tooltip(tooltips.fitting.nfev)
            dpg.add_input_int(
                default_value=state.config.default_fit_settings.max_nfev,
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=update_default_fit_settings,
                tag=fit_max_nfev_input,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=restore_default_fit_settings,
            )
        section_spacer()


def simulation_tab_settings(label_pad: int, state):
    with dpg.collapsing_header(label="Simulation tab", default_open=True):
        sim_cdc_input: int = dpg.generate_uuid()
        sim_max_freq_input: int = dpg.generate_uuid()
        sim_min_freq_input: int = dpg.generate_uuid()
        sim_per_decade_input: int = dpg.generate_uuid()

        def update_default_sim_settings():
            state.config.default_simulation_settings = SimulationSettings(
                dpg.get_value(sim_cdc_input),
                dpg.get_value(sim_min_freq_input),
                dpg.get_value(sim_max_freq_input),
                dpg.get_value(sim_per_decade_input),
            )

        def restore_default_sim_settings():
            dpg.set_value(sim_cdc_input, DEFAULT_SIMULATION_SETTINGS.cdc)
            dpg.set_value(sim_min_freq_input, DEFAULT_SIMULATION_SETTINGS.min_frequency)
            dpg.set_value(sim_max_freq_input, DEFAULT_SIMULATION_SETTINGS.max_frequency)
            dpg.set_value(
                sim_per_decade_input, DEFAULT_SIMULATION_SETTINGS.num_per_decade
            )
            update_default_sim_settings()

        with dpg.group(horizontal=True):
            dpg.add_text("Circuit".rjust(label_pad))
            attach_tooltip(tooltips.simulation.cdc)
            dpg.add_input_text(
                default_value=state.config.default_simulation_settings.cdc,
                on_enter=True,
                callback=update_default_sim_settings,
                width=-1,
                tag=sim_cdc_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Max. frequency".rjust(label_pad))
            attach_tooltip(tooltips.simulation.max_freq)
            dpg.add_input_float(
                label="Hz",
                default_value=state.config.default_simulation_settings.max_frequency,
                min_value=1e-20,
                max_value=1e20,
                min_clamped=True,
                max_clamped=True,
                on_enter=True,
                format="%.3E",
                step=0.0,
                width=-22,
                callback=update_default_sim_settings,
                tag=sim_max_freq_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Min. frequency".rjust(label_pad))
            attach_tooltip(tooltips.simulation.min_freq)
            dpg.add_input_float(
                label="Hz",
                default_value=state.config.default_simulation_settings.min_frequency,
                min_value=1e-20,
                max_value=1e20,
                min_clamped=True,
                max_clamped=True,
                on_enter=True,
                format="%.3E",
                step=0.0,
                width=-22,
                callback=update_default_sim_settings,
                tag=sim_min_freq_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Num. points per decade".rjust(label_pad))
            attach_tooltip(tooltips.simulation.per_decade)
            dpg.add_input_int(
                default_value=state.config.default_simulation_settings.num_per_decade,
                min_value=1,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=update_default_sim_settings,
                tag=sim_per_decade_input,
                width=-1,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=restore_default_sim_settings,
            )
        section_spacer()


def plotting_tab_settings(label_pad: int, state):
    with dpg.collapsing_header(label="Plotting tab - Export plot", default_open=True):
        export_units: int = dpg.generate_uuid()
        export_width: int = dpg.generate_uuid()
        export_height: int = dpg.generate_uuid()
        export_dpi: int = dpg.generate_uuid()
        export_preview: int = dpg.generate_uuid()
        export_title: int = dpg.generate_uuid()
        export_legend: int = dpg.generate_uuid()
        export_legend_location: int = dpg.generate_uuid()
        export_grid: int = dpg.generate_uuid()
        export_tight: int = dpg.generate_uuid()
        export_num_per_decade: int = dpg.generate_uuid()
        export_extension: int = dpg.generate_uuid()
        export_experimental_clear_registry: int = dpg.generate_uuid()
        export_experimental_disable_previews: int = dpg.generate_uuid()

        def restore_default_export_settings():
            state.config.export_units = 1
            state.config.export_width = 10.0
            state.config.export_height = 6.0
            state.config.export_dpi = 100
            state.config.export_preview = 4
            state.config.export_title = True
            state.config.export_legend = True
            state.config.export_legend_location = 0
            state.config.export_grid = False
            state.config.export_tight = False
            state.config.export_num_per_decade = 100
            state.config.export_extension = ".png"
            state.config.export_experimental_clear_registry = True
            state.config.export_experimental_disable_previews = False
            dpg.set_value(export_units, list(UNITS_PER_INCH)[state.config.export_units])
            dpg.set_value(export_width, state.config.export_width)
            dpg.set_value(export_height, state.config.export_height)
            dpg.set_value(export_dpi, state.config.export_dpi)
            dpg.set_value(
                export_preview, list(PREVIEW_LIMITS)[state.config.export_preview]
            )
            dpg.set_value(export_title, state.config.export_title)
            dpg.set_value(export_legend, state.config.export_legend)
            dpg.set_value(
                export_legend_location,
                list(LEGEND_LOCATIONS)[state.config.export_legend_location],
            )
            dpg.set_value(export_grid, state.config.export_grid)
            dpg.set_value(export_tight, state.config.export_tight)
            dpg.set_value(export_num_per_decade, state.config.export_num_per_decade)
            dpg.set_value(export_extension, state.config.export_extension)
            dpg.set_value(
                export_experimental_clear_registry,
                state.config.export_experimental_clear_registry,
            )
            dpg.set_value(
                export_experimental_disable_previews,
                state.config.export_experimental_disable_previews,
            )
            state.plot_exporter.update_settings(state.config)

        def update_export_units(unit: str):
            state.config.export_units = list(UNITS_PER_INCH).index(unit)
            state.plot_exporter.update_settings(state.config)

        def update_export_width(width: float):
            state.config.export_width = width
            state.plot_exporter.update_settings(state.config)

        def update_export_height(height: float):
            state.config.export_height = height
            state.plot_exporter.update_settings(state.config)

        def update_export_dpi(dpi: float):
            state.config.export_dpi = dpi
            state.plot_exporter.update_settings(state.config)

        def update_export_preview(limit: str):
            state.config.export_preview = list(PREVIEW_LIMITS).index(limit)
            state.plot_exporter.update_settings(state.config)

        def update_export_title(flag: bool):
            state.config.export_title = flag
            state.plot_exporter.update_settings(state.config)

        def update_export_legend(flag: bool):
            state.config.export_legend = flag
            state.plot_exporter.update_settings(state.config)

        def update_export_legend_location(location: str):
            state.config.export_legend_location = LEGEND_LOCATIONS[location]
            state.plot_exporter.update_settings(state.config)

        def update_export_grid(flag: bool):
            state.config.export_grid = flag
            state.plot_exporter.update_settings(state.config)

        def update_export_tight(flag: bool):
            state.config.export_tight = flag
            state.plot_exporter.update_settings(state.config)

        def update_export_num_per_decade(num: int):
            state.config.export_num_per_decade = num
            state.plot_exporter.update_settings(state.config)

        def update_export_extension(ext: str):
            state.config.export_extension = EXTENSIONS.index(ext)
            state.plot_exporter.update_settings(state.config)

        def update_export_experimental_clear_registry(flag: bool):
            state.config.export_experimental_clear_registry = flag
            state.plot_exporter.update_settings(state.config)

        def update_export_experimental_disable_previews(flag: bool):
            state.config.export_experimental_disable_previews = flag
            state.plot_exporter.update_settings(state.config)

        with dpg.group(horizontal=True):
            dpg.add_text("Units".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_units)
            dpg.add_combo(
                width=-1,
                items=list(UNITS_PER_INCH.keys()),
                default_value=list(UNITS_PER_INCH.keys())[
                    state.config.export_units
                ],
                callback=lambda s, a, u: update_export_units(a),
                tag=export_units,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Width".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_width)
            dpg.add_input_float(
                width=-1,
                default_value=state.config.export_width,
                min_value=0.01,
                min_clamped=True,
                step=0.0,
                format="%.2f",
                on_enter=True,
                callback=lambda s, a, u: update_export_width(a),
                tag=export_width,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Height".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_height)
            dpg.add_input_float(
                width=-1,
                default_value=state.config.export_height,
                min_value=0.01,
                min_clamped=True,
                step=0.0,
                format="%.2f",
                on_enter=True,
                callback=lambda s, a, u: update_export_height(a),
                tag=export_height,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("DPI".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_dpi)
            dpg.add_input_int(
                width=-1,
                default_value=state.config.export_dpi,
                min_value=1,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=lambda s, a, u: update_export_dpi(a),
                tag=export_dpi,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Preview".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_preview)
            dpg.add_combo(
                width=-1,
                items=list(PREVIEW_LIMITS.keys()),
                default_value=list(PREVIEW_LIMITS.keys())[
                    state.config.export_preview
                ],
                callback=lambda s, a, u: update_export_preview(a),
                tag=export_preview,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Title".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_title)
            dpg.add_checkbox(
                default_value=state.config.export_title,
                callback=lambda s, a, u: update_export_title(a),
                tag=export_title,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Legend".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_legend)
            dpg.add_checkbox(
                default_value=state.config.export_legend,
                callback=lambda s, a, u: update_export_legend(a),
                tag=export_legend,
            )
            dpg.add_combo(
                width=-1,
                items=list(LEGEND_LOCATIONS.keys()),
                default_value=list(LEGEND_LOCATIONS.keys())[0],
                callback=lambda s, a, u: update_export_legend_location(a),
                tag=export_legend_location,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Grid".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_grid)
            dpg.add_checkbox(
                default_value=state.config.export_grid,
                callback=lambda s, a, u: update_export_grid(a),
                tag=export_grid,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Tight".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_tight)
            dpg.add_checkbox(
                default_value=state.config.export_tight,
                callback=lambda s, a, u: update_export_tight(a),
                tag=export_tight,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Points per decade".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_num_per_decade)
            dpg.add_input_int(
                width=-1,
                default_value=state.config.export_num_per_decade,
                min_value=1,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=lambda s, a, u: update_export_num_per_decade(a),
                tag=export_num_per_decade,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Extension".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_extension)
            dpg.add_combo(
                width=-1,
                items=EXTENSIONS,
                default_value=EXTENSIONS[state.config.export_extension],
                callback=lambda s, a, u: update_export_extension(a),
                tag=export_extension,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Clear texture registry".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_experimental_clear_registry)
            dpg.add_checkbox(
                default_value=state.config.export_experimental_clear_registry,
                callback=lambda s, a, u: update_export_experimental_clear_registry(
                    a
                ),
                tag=export_experimental_clear_registry,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Disable plot previews".rjust(label_pad))
            attach_tooltip(tooltips.plotting.export_experimental_disable_previews)
            dpg.add_checkbox(
                default_value=state.config.export_experimental_disable_previews,
                callback=lambda s, a, u: update_export_experimental_disable_previews(
                    a
                ),
                tag=export_experimental_disable_previews,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=restore_default_export_settings,
            )
        section_spacer()


def show_defaults_settings_window(state):
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(374, 540)

    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close_window():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    with dpg.window(
        label="Settings - defaults",
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_resize=True,
        on_close=close_window,
        tag=window,
    ):
        label_pad: int = 22
        general_settings(label_pad, state)
        kramers_kronig_tab_settings(label_pad, state)
        fitting_tab_settings(label_pad, state)
        simulation_tab_settings(label_pad, state)
        plotting_tab_settings(label_pad, state)
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Dict, List, Optional, Tuple
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Plot
from deareis.gui.data_sets import DataSetsTab
from deareis.gui.fitting import FittingTab
from deareis.gui.kramers_kronig import KramersKronigTab
from deareis.gui.overview import OverviewTab
from deareis.gui.plotting import PlottingTab
from deareis.gui.simulation import SimulationTab
from deareis.signals import Signal
import deareis.signals as signals
from deareis.enums import PlotType, Output
from deareis.data import (
    DataSet,
    FitResult,
    FitSettings,
    PlotSettings,
    Project,
    SimulationResult,
    SimulationSettings,
    TestResult,
    TestSettings,
)
import deareis.themes as themes
from deareis.keybindings import Context


class ProjectTab:
    def __init__(self, uuid: str, tab_bar: int, state):
        assert type(uuid) is str, uuid
        assert type(tab_bar) is int, tab_bar
        self.uuid: str = uuid
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(
            label="Project",
            order_mode=dpg.mvTabOrder_Reorderable,
            user_data=uuid,
            tag=self.tab,
            parent=tab_bar,
        ):
            tab_lookup: dict = {}
            self.context_lookup: Dict[int, Context] = {}

            def select_tab(tab_id: int):
                tab = tab_lookup.get(tab_id)
                assert tab is not None
                tab.resize(
                    width=dpg.get_viewport_width(),
                    height=dpg.get_viewport_height(),
                )
                if hasattr(tab, "queued_update") and tab.queued_update is not None:
                    tab.queued_update()

            self.tab_bar: int = dpg.generate_uuid()
            with dpg.tab_bar(
                callback=lambda s, a, u: select_tab(a),
                tag=self.tab_bar,
            ):
                self.overview_tab: OverviewTab = OverviewTab()
                self.data_sets_tab: DataSetsTab = DataSetsTab()
                self.kramers_kronig_tab: KramersKronigTab = KramersKronigTab(state)
                self.fitting_tab: FittingTab = FittingTab(state)
                self.simulation_tab: SimulationTab = SimulationTab(state)
                self.plotting_tab: PlottingTab = PlottingTab(state)
            tab_lookup.update(
                {
                    self.overview_tab.tab: self.overview_tab,
                    self.data_sets_tab.tab: self.data_sets_tab,
                    self.kramers_kronig_tab.tab: self.kramers_kronig_tab,
                    self.fitting_tab.tab: self.fitting_tab,
                    self.simulation_tab.tab: self.simulation_tab,
                    self.plotting_tab.tab: self.plotting_tab,
                }
            )
            self.context_lookup.update(
                {
                    self.overview_tab.tab: Context.OVERVIEW_TAB,
                    self.data_sets_tab.tab: Context.DATA_SETS_TAB,
                    self.kramers_kronig_tab.tab: Context.KRAMERS_KRONIG_TAB,
                    self.fitting_tab.tab: Context.FITTING_TAB,
                    self.simulation_tab.tab: Context.SIMULATION_TAB,
                    self.plotting_tab.tab: Context.PLOTTING_TAB,
                }
            )

            def visibility_handler(tab):
                if not dpg.does_item_exist(tab.tab):
                    return
                elif not tab.is_visible():
                    return
                elif hasattr(tab, "queued_update") and tab.queued_update is not None:
                    tab.queued_update()

            for tab in tab_lookup.values():
                if not (hasattr(tab, "is_visible") and hasattr(tab, "visibility_item")):
                    continue
                item_handler: int = dpg.generate_uuid()
                with dpg.item_handler_registry(tag=item_handler):
                    dpg.add_item_visible_handler(
                        callback=lambda s, a, u: visibility_handler(u),
                        user_data=tab,
                    )
                dpg.bind_item_handler_registry(tab.visibility_item, item_handler)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0, width
        assert type(height) is int and height > 0, height
        self.data_sets_tab.resize(width, height)
        self.kramers_kronig_tab.resize(width, height)
        self.fitting_tab.resize(width, height)
        self.simulation_tab.resize(width, height)
        self.plotting_tab.resize(width, height)

    def select_next_tab(self):
        current_tab: int = dpg.get_value(self.tab_bar)
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        while tabs:
            tab: int = tabs.pop(0)
            if tab == current_tab:
                break
        if not tabs:
            tabs = dpg.get_item_children(self.tab_bar, slot=1)
        tab = tabs[0]
        dpg.set_value(self.tab_bar, tab)

    def select_previous_tab(self):
        current_tab: int = dpg.get_value(self.tab_bar)
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        while tabs:
            tab: int = tabs.pop()
            if tab == current_tab:
                break
        if not tabs:
            tabs = dpg.get_item_children(self.tab_bar, slot=1)
        tab = tabs[-1]
        dpg.set_value(self.tab_bar, tab)

    def get_active_context(self) -> Context:
        return self.context_lookup[dpg.get_value(self.tab_bar)]

    def get_active_data_set(
        self, context: Optional[Context] = None
    ) -> Optional[DataSet]:
        assert type(context) is Context or context is None, context
        tag: Optional[int] = None
        if context is not None:
            tag = {
                Context.DATA_SETS_TAB: self.data_sets_tab.delete_button,
                Context.KRAMERS_KRONIG_TAB: self.kramers_kronig_tab.perform_test_button,
                Context.FITTING_TAB: self.fitting_tab.perform_fit_button,
                Context.SIMULATION_TAB: self.simulation_tab.perform_sim_button,
            }.get(context)
        else:
            tag = {
                self.data_sets_tab.tab: self.data_sets_tab.delete_button,
                self.kramers_kronig_tab.tab: self.kramers_kronig_tab.perform_test_button,
                self.fitting_tab.tab: self.fitting_tab.perform_fit_button,
                self.simulation_tab.tab: self.simulation_tab.perform_sim_button,
            }.get(dpg.get_value(self.tab_bar))
        if tag is None:
            return None
        return dpg.get_item_user_data(tag)

    def get_active_test(self) -> Optional[TestResult]:
        return dpg.get_item_user_data(self.kramers_kronig_tab.delete_button).get("test")

    def get_active_fit(self) -> Optional[FitResult]:
        return dpg.get_item_user_data(self.fitting_tab.delete_button).get("fit")

    def get_active_simulation(self) -> Optional[SimulationResult]:
        return dpg.get_item_user_data(self.simulation_tab.delete_button)

    def get_active_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_active_plot()

    def select_overview_tab(self):
        dpg.set_value(self.tab_bar, self.overview_tab.tab)

    def select_data_sets_tab(self):
        dpg.set_value(self.tab_bar, self.data_sets_tab.tab)

    def select_kramers_kronig_tab(self):
        dpg.set_value(self.tab_bar, self.kramers_kronig_tab.tab)

    def select_fitting_tab(self):
        dpg.set_value(self.tab_bar, self.fitting_tab.tab)

    def select_simulation_tab(self):
        dpg.set_value(self.tab_bar, self.simulation_tab.tab)

    def select_plotting_tab(self):
        dpg.set_value(self.tab_bar, self.plotting_tab.tab)

    def set_label(self, label: str):
        assert type(label) is str and label != "", label
        dpg.set_item_label(self.tab, label)
        self.overview_tab.set_label(label)

    def get_notes(self) -> str:
        return self.overview_tab.get_notes()

    def set_notes(self, notes: str):
        assert type(notes) is str, notes
        self.overview_tab.set_notes(notes)

    def set_dirty(self, state: bool):
        dpg.bind_item_theme(self.tab, themes.tab.dirty if state else themes.tab.clean)

    def populate_data_sets(self, project: Project):
        assert type(project) is Project, project
        data_sets: List[DataSet] = project.get_data_sets()
        lookup: Dict[str, DataSet] = {_.get_label(): _ for _ in data_sets}
        labels: List[str] = list(lookup.keys())
        self.data_sets_tab.populate_data_sets(labels, lookup)
        self.kramers_kronig_tab.populate_data_sets(labels, lookup)
        self.fitting_tab.populate_data_sets(labels, lookup)
        self.simulation_tab.populate_data_sets(labels, lookup)
        settings: Optional[PlotSettings] = self.plotting_tab.get_active_plot()
        if settings is not None:
            self.plotting_tab.populate_data_sets(data_sets, settings)

    def populate_tests(self, project: Project, data: Optional[DataSet]):
        assert type(project) is Project, project
        assert type(data) is DataSet or data is None, data
        self.kramers_kronig_tab.populate_tests(
            {_.get_label(): _ for _ in project.get_tests(data)}
            if data is not None
            else {},
            data,
        )
        settings: Optional[PlotSettings] = self.plotting_tab.get_active_plot()
        if settings is not None:
            self.plotting_tab.populate_tests(
                project.get_all_tests(),
                project.get_data_sets(),
                settings,
            )

    def populate_fits(self, project: Project, data: Optional[DataSet]):
        assert type(project) is Project, project
        assert type(data) is DataSet or data is None, data
        self.fitting_tab.populate_fits(
            {_.get_label(): _ for _ in project.get_fits(data)}
            if data is not None
            else {},
            data,
        )
        settings: Optional[PlotSettings] = self.plotting_tab.get_active_plot()
        if settings is not None:
            self.plotting_tab.populate_fits(
                project.get_all_fits(),
                project.get_data_sets(),
                settings,
            )

    def populate_simulations(self, project: Project):
        assert type(project) is Project, project
        self.simulation_tab.populate_simulations(
            {_.get_label(): _ for _ in project.get_simulations()}
        )
        settings: Optional[PlotSettings] = self.plotting_tab.get_active_plot()
        if settings is not None:
            self.plotting_tab.populate_simulations(
                project.get_simulations(),
                settings,
            )

    def populate_plots(self, project: Project):
        assert type(project) is Project, project
        self.plotting_tab.populate_plots(
            {_.get_label(): _ for _ in project.get_plots()}
        )

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.data_sets_tab.select_data_set(data)
        self.kramers_kronig_tab.select_test_result(None, data)
        self.fitting_tab.select_fit_result(None, data)

    def get_next_data_set(self, context: Context) -> Optional[DataSet]:
        if context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.get_next_data_set()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.get_next_data_set()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.get_next_data_set()
        return None

    def get_previous_data_set(self, context: Context) -> Optional[DataSet]:
        if context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.get_previous_data_set()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.get_previous_data_set()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.get_previous_data_set()
        return None

    def get_next_simulation_data_set(self) -> Optional[DataSet]:
        return self.simulation_tab.get_next_data_set()

    def get_previous_simulation_data_set(self) -> Optional[DataSet]:
        return self.simulation_tab.get_previous_data_set()

    def get_next_test_result(self) -> Optional[TestResult]:
        return self.kramers_kronig_tab.get_next_result()

    def get_previous_test_result(self) -> Optional[TestResult]:
        return self.kramers_kronig_tab.get_previous_result()

    def get_next_fit_result(self) -> Optional[FitResult]:
        return self.fitting_tab.get_next_result()

    def get_previous_fit_result(self) -> Optional[FitResult]:
        return self.fitting_tab.get_previous_result()

    def get_next_simulation_result(self) -> Optional[SimulationResult]:
        return self.simulation_tab.get_next_result()

    def get_previous_simulation_result(self) -> Optional[SimulationResult]:
        return self.simulation_tab.get_previous_result()

    def get_next_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_next_plot()

    def get_previous_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_previous_plot()

    def get_next_plot_type(self) -> Optional[PlotType]:
        return self.plotting_tab.get_next_plot_type()

    def get_previous_plot_type(self) -> Optional[PlotType]:
        return self.plotting_tab.get_previous_plot_type()

    def select_test_result(self, test: TestResult, data: DataSet):
        assert type(test) is TestResult, test
        assert type(data) is DataSet, data
        self.kramers_kronig_tab.select_test_result(test, data)

    def select_fit_result(self, fit: FitResult, data: DataSet):
        assert type(fit) is FitResult, fit
        assert type(data) is DataSet, data
        self.fitting_tab.select_fit_result(fit, data)

    def select_simulation_result(
        self, simulation: Optional[SimulationResult], data: Optional[DataSet]
    ):
        assert type(simulation) is SimulationResult or simulation is None, simulation
        assert type(data) is DataSet or data is None, data
        self.simulation_tab.select_simulation_result(simulation, data)

    def select_plot(
        self,
        settings: PlotSettings,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        adjust_limits: bool,
        plot_only: bool,
    ):
        assert type(settings) is PlotSettings, settings
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(adjust_limits) is bool, adjust_limits
        assert type(plot_only) is bool, plot_only
        self.plotting_tab.select_plot(
            settings, data_sets, tests, fits, simulations, adjust_limits, plot_only
        )

    def select_plot_type(self, plot_type: PlotType):
        assert type(plot_type) is PlotType, plot_type
        self.plotting_tab.select_plot_type(plot_type)

    def set_series_label(self, uuid: str, label: Optional[str]):
        assert type(uuid) is str, uuid
        assert type(label) is str or label is None, label
        self.plotting_tab.set_series_label(uuid, label)

    def get_test_settings(self) -> TestSettings:
        return self.kramers_kronig_tab.get_settings()

    def set_test_settings(self, settings: TestSettings):
        assert type(settings) is TestSettings, settings
        self.kramers_kronig_tab.set_settings(settings)

    def get_fit_settings(self) -> FitSettings:
        return self.fitting_tab.get_settings()

    def set_fit_settings(self, settings: FitSettings):
        assert type(settings) is FitSettings, settings
        self.fitting_tab.set_settings(settings)

    def get_simulation_settings(self) -> SimulationSettings:
        return self.simulation_tab.get_settings()

    def set_simulation_settings(self, settings: SimulationSettings):
        assert type(settings) is SimulationSettings, settings
        self.simulation_tab.set_settings(settings)

    def show_enlarged_nyquist(self):
        {
            self.data_sets_tab.tab: self.data_sets_tab.show_enlarged_nyquist,
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_nyquist,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_nyquist,
            self.simulation_tab.tab: self.simulation_tab.show_enlarged_nyquist,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_bode(self):
        {
            self.data_sets_tab.tab: self.data_sets_tab.show_enlarged_bode,
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_bode,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_bode,
            self.simulation_tab.tab: self.simulation_tab.show_enlarged_bode,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_residuals(self):
        {
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_residuals,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_residuals,
        }.get(dpg.get_value(self.tab_bar))()

    def update_plots(
        self,
        settings: PlotSettings,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
    ):
        assert type(settings) is PlotSettings, settings
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        self.plotting_tab.plot_series(
            data_sets,
            tests,
            fits,
            simulations,
            settings,
            adjust_limits=False,
        )

    def get_nyquist_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.nyquist_plot
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.nyquist_plot
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.nyquist_plot
        return None

    def get_bode_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.bode_plot_horizontal
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.bode_plot_horizontal
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.bode_plot_horizontal
        return None

    def get_residuals_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.residuals_plot
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.residuals_plot
        return None

    def get_active_output(self, context: Context) -> Optional[Output]:
        if context == Context.FITTING_TAB:
            return self.fitting_tab.get_active_output()
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.get_active_output()
        return None

    def get_filtered_plot_series(
        self,
    ) -> Tuple[
        List[DataSet], List[TestResult], List[FitResult], List[SimulationResult]
    ]:
        string: str = self.plotting_tab.get_filter_string()
        return (
            self.plotting_tab.possible_data_sets.filter(string, False),
            self.plotting_tab.possible_tests.filter(string, False),
            self.plotting_tab.possible_fits.filter(string, False),
            self.plotting_tab.possible_simulations.filter(string, False),
        )

    def has_active_input(self, context: Optional[Context] = None) -> bool:
        assert type(context) is Context or context is None, context
        if context is None:
            context = self.get_active_context()
        if context == Context.OVERVIEW_TAB:
            return self.overview_tab.has_active_input()
        elif context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.has_active_input()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.has_active_input()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.has_active_input()
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.has_active_input()
        elif context == Context.PLOTTING_TAB:
            return self.plotting_tab.has_active_input()
        return False



from setuptools import (
    setup,
    find_packages,
)
from os import walk
from os.path import (
    basename,
    exists,
    join,
)


entry_points = {
    "gui_scripts": [
        "deareis = deareis.program:main",
    ],
    "console_scripts": [
        "deareis-debug = deareis.program:debug",  # For the convenience of users on Windows
    ],
}


# The version number defined below is propagated to /src/deareis/version.py
# when running this script.
version = "2.1.0"


licenses = []
for _, _, files in walk("LICENSES"):
    licenses.extend(
        list(
            map(
                lambda _: join("LICENSES", _),
                filter(lambda _: _.startswith("LICENSE-"), files),
            )
        )
    )


def update_file(src: str, dst: str):
    if not exists(src):
        return
    src_contents = ""
    with open(src, "r") as fp:
        src_contents = fp.read()
    if exists(dst):
        with open(dst, "r") as fp:
            if fp.read() == src_contents:
                return
    with open(dst, "w") as fp:
        fp.write(src_contents)


copyright_notice = ""
with open("COPYRIGHT") as fp:
    copyright_notice = fp.read().strip()

with open(join("src", "deareis", "version.py"), "w") as fp:
    fp.write(f'{copyright_notice}\n\nPACKAGE_VERSION: str = "{version}"')

# The changelog bundled with the package will also be updated when running this script.
update_file(
    join("CHANGELOG.md"),
    join("src", "deareis", "gui", "changelog", "CHANGELOG.md"),
)

# The licenses bundled with the package will also be updated when running this script.
update_file(
    join("LICENSE"),
    join("src", "deareis", "gui", "licenses", "LICENSE-DearEIS.txt"),
)
list(
    map(
        lambda _: update_file(
            _,
            join("src", "deareis", "gui", "licenses", basename(_)),
        ),
        licenses,
    )
)


setup(
    name="deareis",
    version=version,
    author="DearEIS developers",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    data_files=[
        "COPYRIGHT",
        "CONTRIBUTORS",
        "LICENSES/README.md",
        "src/deareis/gui/changelog/CHANGELOG.md",
    ]
    + licenses,
    url="https://vyrjana.github.io/DearEIS",
    project_urls={
        "Documentation": "https://vyrjana.github.io/DearEIS/api/",
        "Source Code": "https://github.com/vyrjana/DearEIS",
        "Bug Tracker": "https://github.com/vyrjana/DearEIS/issues",
    },
    license="GPLv3",
    description="A GUI program for analyzing, simulating, and visualizing impedance spectra.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    entry_points=entry_points,
    install_requires=[
        "dearpygui>=1.6.2",  # Used to implement the GUI.
        "pyimpspec>=2.1.0",  # Used for parsing, fitting, and analyzing impedance spectra.
        "requests>=2.27.1",  # Used to check package status on PyPI.
        "tabulate>=0.8.10",  # Required by pandas to generate Markdown tables.
        "xdg>=5.1.1",  # Used to figure out where to place config, state, etc. files.
    ],
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering",
    ],
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
import deareis.themes as themes
from deareis.themes import PLOT_MARKERS
from typing import Dict, List, Optional, Tuple
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
from deareis.state import STATE
from deareis.gui.plots import (
    Bode,
    Nyquist,
    Residuals,
    MuXps,
)
from numpy import array, logspace, log10 as log, ndarray
from numpy.random import normal
from pyimpspec import Circuit
from deareis.data import DataSet
import pyimpspec
from deareis.signals import Signal
import deareis.signals as signals


class AppearanceSettings:
    def __init__(self):
        self.marker_items: List[str] = list(PLOT_MARKERS.keys())
        self.marker_label_lookup: Dict[int, str] = {
            v: k for k, v in PLOT_MARKERS.items()
        }
        self.label_pad: int = 56
        self.data: DataSet
        self.sim_data: DataSet
        self.smooth_data: DataSet
        self.real_residual: ndarray
        self.imag_residual: ndarray
        self.noise: ndarray
        (
            self.data,
            self.sim_data,
            self.smooth_data,
            self.real_residual,
            self.imag_residual,
            self.noise,
        ) = self.generate_data()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(600, 540)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Settings - appearance",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            no_move=False,
            no_resize=True,
            on_close=self.close_window,
            tag=self.window,
        ):
            with dpg.collapsing_header(label="General", default_open=True):
                with dpg.group(horizontal=True):
                    dpg.add_text(
                        "Number of points per decade in simulated response".rjust(
                            self.label_pad
                        )
                    )
                    attach_tooltip(
                        """
This affects how smooth the lines will look but it may also affect performance when rendering the graphical user interface. This setting also affects how many points are included when copying plot data as character-separated values (CSV). Changes made to this setting will take effect the next time a plot is redrawn.
                        """.strip()
                    )
                    dpg.add_slider_int(
                        default_value=STATE.config.num_per_decade_in_simulated_lines,
                        min_value=1,
                        max_value=200,
                        clamped=True,
                        callback=self.update_simulated_num_per_decade,
                        width=-1,
                    )
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Bode plots", default_open=True):
                self.bode_plot = Bode(width=-1, height=200)
                # bode_show_legend_checkbox: int = dpg.generate_uuid()
                # bode_legend_outside_checkbox: int = dpg.generate_uuid()
                # bode_horizontal_legend_checkbox: int = dpg.generate_uuid()
                # bode_legend_location_combo: int = dpg.generate_uuid()
                self.bode_data_mag_color: int = dpg.generate_uuid()
                self.bode_data_phase_color: int = dpg.generate_uuid()
                self.bode_sim_mag_color: int = dpg.generate_uuid()
                self.bode_sim_phase_color: int = dpg.generate_uuid()
                self.bode_data_mag_marker: int = dpg.generate_uuid()
                self.bode_data_phase_marker: int = dpg.generate_uuid()
                self.bode_sim_mag_marker: int = dpg.generate_uuid()
                self.bode_sim_phase_marker: int = dpg.generate_uuid()
                # Legend checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Show legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_bode_plot)
                # Legend outside
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend outside plot".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_bode_plot)
                # Legend horizontal checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Horizontal legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_bode_plot)
                # Legend location combo
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend location".rjust(label_pad))
                #    dpg.add_combo(width=-1, callback=update_bode_plot)
                # Data colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - magnitude".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_magnitude_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.magnitude_data,
                        tag=self.bode_data_mag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_magnitude_data"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.magnitude_data,
                        tag=self.bode_data_mag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Data - phase".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_phase_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.phase_data,
                        tag=self.bode_data_phase_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_phase_data"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.phase_data,
                        tag=self.bode_data_phase_marker,
                        width=-1,
                    )
                # Sim/fit colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - magnitude".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_magnitude_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.magnitude_simulation,
                        tag=self.bode_sim_mag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_magnitude_simulation"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.magnitude_simulation,
                        tag=self.bode_sim_mag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation - phase".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["bode_phase_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_bode_color,
                        user_data=themes.bode.phase_simulation,
                        tag=self.bode_sim_phase_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["bode_phase_simulation"]
                        ],
                        callback=self.update_bode_marker,
                        user_data=themes.bode.phase_simulation,
                        tag=self.bode_sim_phase_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_bode_plot,
                        width=-1,
                    )
                self.update_bode_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Nyquist plots", default_open=True):
                self.nyquist_plot = Nyquist(width=-1, height=200)
                # nyquist_show_legend_checkbox: int = dpg.generate_uuid()
                # nyquist_legend_outside_checkbox: int = dpg.generate_uuid()
                # nyquist_horizontal_legend_checkbox: int = dpg.generate_uuid()
                # nyquist_legend_location_combo: int = dpg.generate_uuid()
                self.nyquist_data_color: int = dpg.generate_uuid()
                self.nyquist_sim_color: int = dpg.generate_uuid()
                self.nyquist_data_marker: int = dpg.generate_uuid()
                self.nyquist_sim_marker: int = dpg.generate_uuid()
                # Legend checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Show legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_nyquist_plot)
                # Legend outside
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend outside plot".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_nyquist_plot)
                # Legend horizontal checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Horizontal legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_nyquist_plot)
                # Legend location combo
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend location".rjust(label_pad))
                #    dpg.add_combo(width=-1, callback=update_nyquist_plot)
                # Data colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Data".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["nyquist_data"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_nyquist_color,
                        user_data=themes.nyquist.data,
                        tag=self.nyquist_data_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["nyquist_data"]
                        ],
                        callback=self.update_nyquist_marker,
                        user_data=themes.nyquist.data,
                        tag=self.nyquist_data_marker,
                        width=-1,
                    )
                # Sim/fit colors and markers
                with dpg.group(horizontal=True):
                    dpg.add_text("Fit/simulation".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["nyquist_simulation"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_nyquist_color,
                        user_data=themes.nyquist.simulation,
                        tag=self.nyquist_sim_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["nyquist_simulation"]
                        ],
                        callback=self.update_nyquist_marker,
                        user_data=themes.nyquist.simulation,
                        tag=self.nyquist_sim_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_nyquist_plot,
                        width=-1,
                    )
                self.update_nyquist_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="Residuals plots", default_open=True):
                self.residuals_plot = Residuals(width=-1, height=200)
                # residuals_show_legend_checkbox: int = dpg.generate_uuid()
                # residuals_legend_outside_checkbox: int = dpg.generate_uuid()
                # residuals_horizontal_legend_checkbox: int = dpg.generate_uuid()
                # residuals_legend_location_combo: int = dpg.generate_uuid()
                self.residuals_real_color: int = dpg.generate_uuid()
                self.residuals_imag_color: int = dpg.generate_uuid()
                self.residuals_real_marker: int = dpg.generate_uuid()
                self.residuals_imag_marker: int = dpg.generate_uuid()
                # Legend checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Show legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_residuals_plot)
                # Legend outside
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend outside plot".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_residuals_plot)
                # Legend horizontal checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Horizontal legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_residuals_plot)
                # Legend location combo
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend location".rjust(label_pad))
                #    dpg.add_combo(width=-1, callback=update_residuals_plot)
                # Zre color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("Z' error".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["residuals_real"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_residuals_color,
                        user_data=themes.residuals.real,
                        tag=self.residuals_real_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["residuals_real"]
                        ],
                        callback=self.update_residuals_marker,
                        user_data=themes.residuals.real,
                        tag=self.residuals_real_marker,
                        width=-1,
                    )
                # Zim color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text('Z" error'.rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["residuals_imaginary"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_residuals_color,
                        user_data=themes.residuals.imaginary,
                        tag=self.residuals_imag_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["residuals_imaginary"]
                        ],
                        callback=self.update_residuals_marker,
                        user_data=themes.residuals.imaginary,
                        tag=self.residuals_imag_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_residuals_plot,
                        width=-1,
                    )
                self.update_residuals_plot(True)
                dpg.add_spacer(height=8)
            with dpg.collapsing_header(label="µ-X² (pseudo) plots", default_open=True):
                self.muxps_plot = MuXps(width=-1, height=200)
                # muxps_show_legend_checkbox: int = dpg.generate_uuid()
                # muxps_legend_outside_checkbox: int = dpg.generate_uuid()
                # muxps_horizontal_legend_checkbox: int = dpg.generate_uuid()
                # muxps_legend_location_combo: int = dpg.generate_uuid()
                self.muxps_mu_criterion_color: int = dpg.generate_uuid()
                self.muxps_mu_color: int = dpg.generate_uuid()
                self.muxps_mu_highlight_color: int = dpg.generate_uuid()
                self.muxps_xps_color: int = dpg.generate_uuid()
                self.muxps_xps_highlight_color: int = dpg.generate_uuid()
                self.muxps_mu_marker: int = dpg.generate_uuid()
                self.muxps_xps_marker: int = dpg.generate_uuid()
                # Legend checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Show legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_muxps_plot)
                # Legend outside
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend outside plot".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_muxps_plot)
                # Legend horizontal checkbox
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Horizontal legend".rjust(label_pad))
                #    dpg.add_checkbox(callback=update_muxps_plot)
                # Legend location combo
                # with dpg.group(horizontal=True):
                #    dpg.add_text("Legend location".rjust(label_pad))
                #    dpg.add_combo(width=-1, callback=update_muxps_plot)
                # Mu-criterion color
                with dpg.group(horizontal=True):
                    dpg.add_text("µ-criterion".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu_criterion"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu_criterion,
                        tag=self.muxps_mu_criterion_color,
                    )
                # Mu color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("µ".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu,
                        tag=self.muxps_mu_color,
                    )
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_mu_highlight"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.mu_highlight,
                        tag=self.muxps_mu_highlight_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["mu_Xps_mu"]
                        ],
                        callback=self.update_muxps_marker,
                        user_data=themes.mu_Xps.mu,
                        tag=self.muxps_mu_marker,
                        width=-1,
                    )
                # Xps color and marker
                with dpg.group(horizontal=True):
                    dpg.add_text("X² (pseudo)".rjust(self.label_pad))
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_Xps"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.Xps,
                        tag=self.muxps_xps_color,
                    )
                    dpg.add_color_edit(
                        default_value=STATE.config.colors["mu_Xps_Xps_highlight"],
                        alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                        no_inputs=True,
                        alpha_bar=True,
                        callback=self.update_muxps_color,
                        user_data=themes.mu_Xps.Xps_highlight,
                        tag=self.muxps_xps_highlight_color,
                    )
                    dpg.add_combo(
                        items=self.marker_items,
                        default_value=self.marker_label_lookup[
                            STATE.config.markers["mu_Xps_Xps"]
                        ],
                        callback=self.update_muxps_marker,
                        user_data=themes.mu_Xps.Xps,
                        tag=self.muxps_xps_marker,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(self.label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=self.reset_muxps_plot,
                        width=-1,
                    )
                self.update_muxps_plot()
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close_window,
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window)

    def close_window(self):
        if dpg.does_item_exist(self.window):
            dpg.delete_item(self.window)
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def generate_data(
        self,
        noise: Optional[ndarray] = None,
    ) -> Tuple[DataSet, DataSet, DataSet, ndarray, ndarray, ndarray]:
        circuit: Circuit = pyimpspec.string_to_circuit(
            "R{R=100}(C{C=1e-6}[R{R=500}W{Y=4e-4}])"
        )
        data: DataSet = pyimpspec.simulate_spectrum(circuit, logspace(0, 5, num=20))
        Z: complex
        sd: float = 0.01
        if noise is None:
            noise = array(
                list(
                    map(
                        lambda Z: complex(
                            abs(Z) * normal(0, sd, 1),
                            abs(Z) * normal(0, sd, 1),
                        ),
                        data.get_impedance(),
                    )
                )
            )
        data.subtract_impedance(-noise)
        sim_data: DataSet = pyimpspec.simulate_spectrum(circuit, data.get_frequency())
        smooth_data: DataSet = pyimpspec.simulate_spectrum(
            circuit,
            logspace(0, 5, num=5 * STATE.config.num_per_decade_in_simulated_lines + 1),
        )
        real_residual: ndarray = (
            (data.get_impedance().real - sim_data.get_impedance().real)
            / abs(data.get_impedance())
            * 100
        )
        imag_residual: ndarray = (
            (data.get_impedance().imag - sim_data.get_impedance().imag)
            / abs(data.get_impedance())
            * 100
        )
        return (data, sim_data, smooth_data, real_residual, imag_residual, noise)

    def update_simulated_num_per_decade(self, sender: int, value: int):
        STATE.config.num_per_decade_in_simulated_lines = value
        (
            self.data,
            self.sim_data,
            self.smooth_data,
            self.real_residual,
            self.imag_residual,
            _,
        ) = self.generate_data(self.noise)
        self.update_bode_plot()
        self.update_nyquist_plot()
        self.update_residuals_plot()
        self.update_muxps_plot()

    def update_bode_plot(self, adjust_limits: bool = False):
        self.bode_plot.clear()
        freq, mag, phase = self.data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (d)",
                "phi (d)",
            ),
            themes=(
                themes.bode.magnitude_data,
                themes.bode.phase_data,
            ),
        )
        freq, mag, phase = self.smooth_data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (fit/sim.)",
                "phi (fit/sim.)",
            ),
            line=True,
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        freq, mag, phase = self.sim_data.get_bode_data()
        self.bode_plot.plot(
            frequency=freq,
            magnitude=mag,
            phase=phase,
            labels=(
                "|Z| (fit/sim.)",
                "phi (fit/sim.)",
            ),
            line=False,
            show_labels=False,
            themes=(
                themes.bode.magnitude_simulation,
                themes.bode.phase_simulation,
            ),
        )
        if adjust_limits:
            self.bode_plot.adjust_limits()

    def update_bode_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.bode.magnitude_data: "bode_magnitude_data",
                themes.bode.phase_data: "bode_phase_data",
                themes.bode.magnitude_simulation: "bode_magnitude_simulation",
                themes.bode.phase_simulation: "bode_phase_simulation",
            }[theme]
        ] = color

    def update_bode_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.bode.magnitude_data: "bode_magnitude_data",
                themes.bode.phase_data: "bode_phase_data",
                themes.bode.magnitude_simulation: "bode_magnitude_simulation",
                themes.bode.phase_simulation: "bode_phase_simulation",
            }[theme]
        ] = marker

    def reset_bode_plot(self):
        dpg.set_value(
            self.bode_data_mag_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.bode_data_phase_color,
            (
                238,
                119,
                51,
                190,
            ),
        )
        dpg.set_value(
            self.bode_sim_mag_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.bode_sim_phase_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        self.update_bode_color(
            self.bode_data_mag_color, None, themes.bode.magnitude_data
        )
        self.update_bode_color(self.bode_data_phase_color, None, themes.bode.phase_data)
        self.update_bode_color(
            self.bode_sim_mag_color, None, themes.bode.magnitude_simulation
        )
        self.update_bode_color(
            self.bode_sim_phase_color, None, themes.bode.phase_simulation
        )
        dpg.set_value(self.bode_data_mag_marker, "Circle")
        dpg.set_value(self.bode_data_phase_marker, "Square")
        dpg.set_value(self.bode_sim_mag_marker, "Cross")
        dpg.set_value(self.bode_sim_phase_marker, "Plus")
        self.update_bode_marker(
            self.bode_data_mag_marker,
            dpg.get_value(self.bode_data_mag_marker),
            themes.bode.magnitude_data,
        )
        self.update_bode_marker(
            self.bode_data_phase_marker,
            dpg.get_value(self.bode_data_phase_marker),
            themes.bode.phase_data,
        )
        self.update_bode_marker(
            self.bode_sim_mag_marker,
            dpg.get_value(self.bode_sim_mag_marker),
            themes.bode.magnitude_simulation,
        )
        self.update_bode_marker(
            self.bode_sim_phase_marker,
            dpg.get_value(self.bode_sim_phase_marker),
            themes.bode.phase_simulation,
        )

    def update_nyquist_plot(self, adjust_limits: bool = False):
        self.nyquist_plot.clear()
        real, imag = self.data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Data",
            theme=themes.nyquist.data,
        )
        real, imag = self.smooth_data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit/sim.",
            line=True,
            theme=themes.nyquist.simulation,
        )
        real, imag = self.sim_data.get_nyquist_data()
        self.nyquist_plot.plot(
            real=real,
            imaginary=imag,
            label="Fit/sim.",
            show_label=False,
            theme=themes.nyquist.simulation,
        )
        if adjust_limits:
            self.nyquist_plot.adjust_limits()

    def update_nyquist_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.nyquist.data: "nyquist_data",
                themes.nyquist.simulation: "nyquist_simulation",
            }[theme]
        ] = color

    def update_nyquist_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.nyquist.data: "nyquist_data",
                themes.nyquist.simulation: "nyquist_simulation",
            }[theme]
        ] = marker

    def reset_nyquist_plot(self):
        dpg.set_value(
            self.nyquist_data_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.nyquist_sim_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        self.update_nyquist_color(self.nyquist_data_color, None, themes.nyquist.data)
        self.update_nyquist_color(
            self.nyquist_sim_color, None, themes.nyquist.simulation
        )
        dpg.set_value(self.nyquist_data_marker, "Circle")
        dpg.set_value(self.nyquist_sim_marker, "Cross")
        self.update_nyquist_marker(
            self.nyquist_data_marker,
            dpg.get_value(self.nyquist_data_marker),
            themes.nyquist.data,
        )
        self.update_nyquist_marker(
            self.nyquist_sim_marker,
            dpg.get_value(self.nyquist_sim_marker),
            themes.nyquist.simulation,
        )

    def update_residuals_plot(self, adjust_limits: bool = False):
        self.residuals_plot.clear()
        self.residuals_plot.plot(
            frequency=log(self.data.get_frequency()),
            real=self.real_residual,
            imaginary=self.imag_residual,
        )
        if adjust_limits:
            self.residuals_plot.adjust_limits()

    def update_residuals_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.residuals.real: "residuals_real",
                themes.residuals.imaginary: "residuals_imaginary",
            }[theme]
        ] = color

    def update_residuals_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.residuals.real: "residuals_real",
                themes.residuals.imaginary: "residuals_imaginary",
            }[theme]
        ] = marker

    def reset_residuals_plot(self):
        dpg.set_value(
            self.residuals_real_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.residuals_imag_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        self.update_residuals_color(
            self.residuals_real_color, None, themes.residuals.real
        )
        self.update_residuals_color(
            self.residuals_imag_color, None, themes.residuals.imaginary
        )
        dpg.set_value(self.residuals_real_marker, "Circle")
        dpg.set_value(self.residuals_imag_marker, "Square")
        self.update_residuals_marker(
            self.residuals_real_marker,
            dpg.get_value(self.residuals_real_marker),
            themes.residuals.real,
        )
        self.update_residuals_marker(
            self.residuals_imag_marker,
            dpg.get_value(self.residuals_imag_marker),
            themes.residuals.imaginary,
        )

    def update_muxps_plot(self):
        self.muxps_plot.clear()
        self.muxps_plot.plot(
            num_RCs=array([1, 2, 3, 4, 5, 6]),
            mu=array([1.0, 0.8, 0.6, 0.4, 0.2, 0.0]),
            Xps=array([0, -1, -1.5, -1.8, -2, -1.4]),
            mu_criterion=0.7,
            num_RC=3,
        )
        self.muxps_plot.adjust_limits()

    def update_muxps_color(self, sender: int, _, theme: int):
        assert type(sender) is int
        assert type(theme) is int
        color: List[float] = dpg.get_value(sender)
        themes.update_plot_series_theme_color(theme, color)
        STATE.config.colors[
            {
                themes.mu_Xps.mu_criterion: "mu_Xps_mu_criterion",
                themes.mu_Xps.mu: "mu_Xps_mu",
                themes.mu_Xps.mu_highlight: "mu_Xps_mu_highlight",
                themes.mu_Xps.Xps: "mu_Xps_Xps",
                themes.mu_Xps.Xps_highlight: "mu_Xps_Xps_highlight",
            }[theme]
        ] = color

    def update_muxps_marker(self, sender: int, label: str, theme: int):
        assert type(sender) is int
        assert type(label) is str
        assert type(theme) is int
        marker: int = PLOT_MARKERS[label]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.mu_Xps.mu: "mu_Xps_mu",
                themes.mu_Xps.Xps: "mu_Xps_Xps",
            }[theme]
        ] = marker
        theme = {
            themes.mu_Xps.mu: themes.mu_Xps.mu_highlight,
            themes.mu_Xps.Xps: themes.mu_Xps.Xps_highlight,
        }[theme]
        themes.update_plot_series_theme_marker(theme, marker)
        STATE.config.markers[
            {
                themes.mu_Xps.mu_highlight: "mu_Xps_mu_highlight",
                themes.mu_Xps.Xps_highlight: "mu_Xps_Xps_highlight",
            }[theme]
        ] = marker

    def reset_muxps_plot(self):
        dpg.set_value(
            self.muxps_mu_criterion_color,
            (
                255,
                255,
                255,
                128,
            ),
        )
        dpg.set_value(
            self.muxps_mu_color,
            (
                238,
                51,
                119,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_mu_highlight_color,
            (
                51,
                187,
                238,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_xps_color,
            (
                0,
                153,
                136,
                190,
            ),
        )
        dpg.set_value(
            self.muxps_xps_highlight_color,
            (
                238,
                119,
                51,
                190,
            ),
        )
        self.update_muxps_color(
            self.muxps_mu_criterion_color, None, themes.mu_Xps.mu_criterion
        )
        self.update_muxps_color(self.muxps_mu_color, None, themes.mu_Xps.mu)
        self.update_muxps_color(
            self.muxps_mu_highlight_color, None, themes.mu_Xps.mu_highlight
        )
        self.update_muxps_color(self.muxps_xps_color, None, themes.mu_Xps.Xps)
        self.update_muxps_color(
            self.muxps_xps_highlight_color,
            None,
            themes.mu_Xps.Xps_highlight,
        )
        dpg.set_value(self.muxps_mu_marker, "Circle")
        dpg.set_value(self.muxps_xps_marker, "Square")
        self.update_muxps_marker(
            self.muxps_mu_marker,
            dpg.get_value(self.muxps_mu_marker),
            themes.mu_Xps.mu,
        )
        self.update_muxps_marker(
            self.muxps_xps_marker,
            dpg.get_value(self.muxps_xps_marker),
            themes.mu_Xps.Xps,
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from json import (
    dumps as dump_json,
    load as load_json,
    loads as parse_json,
)
from os import makedirs
from os.path import (
    exists,
    isdir,
    join,
)
from traceback import format_exc
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    List,
    Optional,
    Set,
)
from xdg import (
    xdg_cache_home,  #  User-specific cache files
    xdg_config_home,  # User-specific configuration files
    xdg_data_home,  #   User-specific data files
)
import dearpygui.dearpygui as dpg
from deareis.keybindings import Keybinding
from deareis.data import (
    FitSettings,
    SimulationSettings,
    TestSettings,
)
from deareis.enums import (
    Action,
    Context,
    Method,
    Mode,
    Test,
    Weight,
    action_contexts,
)
import deareis.themes as themes
from deareis.signals import Signal
import deareis.signals as signals


DEFAULT_KEYBINDINGS: List[Keybinding] = [
    Keybinding(
        dpg.mvKey_N,
        False,
        True,
        False,
        Action.NEW_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_O,
        False,
        True,
        False,
        Action.LOAD_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_Q,
        False,
        True,
        False,
        Action.EXIT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        False,
        True,
        Action.NEXT_PROGRAM_TAB,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        False,
        True,
        Action.PREVIOUS_PROGRAM_TAB,
    ),
    Keybinding(
        dpg.mvKey_F1,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_APPEARANCE,
    ),
    Keybinding(
        dpg.mvKey_F2,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_DEFAULTS,
    ),
    Keybinding(
        dpg.mvKey_F3,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_KEYBINDINGS,
    ),
    Keybinding(
        dpg.mvKey_F11,
        False,
        False,
        False,
        Action.SHOW_HELP_LICENSES,
    ),
    Keybinding(
        dpg.mvKey_F12,
        False,
        False,
        False,
        Action.SHOW_HELP_ABOUT,
    ),
    Keybinding(
        dpg.mvKey_P,
        False,
        True,
        False,
        Action.SHOW_COMMAND_PALETTE,
    ),
    Keybinding(
        dpg.mvKey_S,
        False,
        True,
        False,
        Action.SAVE_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_S,
        False,
        True,
        True,
        Action.SAVE_PROJECT_AS,
    ),
    Keybinding(
        dpg.mvKey_W,
        False,
        True,
        False,
        Action.CLOSE_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_Z,
        False,
        True,
        False,
        Action.UNDO,
    ),
    Keybinding(
        dpg.mvKey_Y if dpg.get_platform() == dpg.mvPlatform_Windows else dpg.mvKey_Z,
        False,
        True,
        False if dpg.get_platform() == dpg.mvPlatform_Windows else True,
        Action.REDO,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        True,
        False,
        Action.NEXT_PROJECT_TAB,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        True,
        False,
        Action.PREVIOUS_PROJECT_TAB,
    ),
    Keybinding(
        dpg.mvKey_D,
        True,
        True,
        False,
        Action.SELECT_DATA_SETS_TAB,
    ),
    Keybinding(
        dpg.mvKey_F,
        True,
        True,
        False,
        Action.SELECT_FITTING_TAB,
    ),
    Keybinding(
        dpg.mvKey_K,
        True,
        True,
        False,
        Action.SELECT_KRAMERS_KRONIG_TAB,
    ),
    Keybinding(
        dpg.mvKey_O,
        True,
        True,
        False,
        Action.SELECT_OVERVIEW_TAB,
    ),
    Keybinding(
        dpg.mvKey_P,
        True,
        True,
        False,
        Action.SELECT_PLOTTING_TAB,
    ),
    Keybinding(
        dpg.mvKey_S,
        True,
        True,
        False,
        Action.SELECT_SIMULATION_TAB,
    ),
    Keybinding(
        dpg.mvKey_Return,
        False if dpg.get_platform() == dpg.mvPlatform_Windows else True,
        True if dpg.get_platform() == dpg.mvPlatform_Windows else False,
        False,
        Action.PERFORM_ACTION,
    ),
    Keybinding(
        dpg.mvKey_Delete,
        True,
        False,
        False,
        Action.DELETE_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        False,
        False,
        Action.NEXT_PRIMARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        False,
        False,
        Action.PREVIOUS_PRIMARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        True,
        False,
        False,
        Action.NEXT_SECONDARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        True,
        False,
        False,
        Action.PREVIOUS_SECONDARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.APPLY_SETTINGS,
    ),
    Keybinding(
        dpg.mvKey_M,
        True,
        False,
        False,
        Action.APPLY_MASK,
    ),
    Keybinding(
        dpg.mvKey_N,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_NYQUIST,
    ),
    Keybinding(
        dpg.mvKey_B,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_BODE,
    ),
    Keybinding(
        dpg.mvKey_R,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_RESIDUALS,
    ),
    Keybinding(
        dpg.mvKey_E,
        True,
        False,
        False,
        Action.SHOW_CIRCUIT_EDITOR,
    ),
    Keybinding(
        dpg.mvKey_N,
        True,
        False,
        True,
        Action.COPY_NYQUIST_DATA,
    ),
    Keybinding(
        dpg.mvKey_B,
        True,
        False,
        True,
        Action.COPY_BODE_DATA,
    ),
    Keybinding(
        dpg.mvKey_R,
        True,
        False,
        True,
        Action.COPY_RESIDUALS_DATA,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_OUTPUT,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.AVERAGE_DATA_SETS,
    ),
    Keybinding(
        dpg.mvKey_T,
        True,
        False,
        False,
        Action.TOGGLE_DATA_POINTS,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_DATA_SET_MASK,
    ),
    Keybinding(
        dpg.mvKey_S,
        True,
        False,
        False,
        Action.SUBTRACT_IMPEDANCE,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.SELECT_ALL_PLOT_SERIES,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        True,
        Action.UNSELECT_ALL_PLOT_SERIES,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_PLOT_APPEARANCE,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        True,
        Action.COPY_PLOT_DATA,
    ),
    Keybinding(
        dpg.mvKey_E,
        True,
        False,
        False,
        Action.EXPAND_COLLAPSE_SIDEBAR,
    ),
    Keybinding(
        dpg.mvKey_P,
        True,
        False,
        False,
        Action.EXPORT_PLOT,
    ),
]


# TODO: Replace string keys with int keys (e.g. themes.nyquist.data) and use strings only in the
# config file
DEFAULT_MARKERS: Dict[str, int] = {
    "residuals_real": dpg.mvPlotMarker_Circle,
    "residuals_imaginary": dpg.mvPlotMarker_Square,
    "nyquist_data": dpg.mvPlotMarker_Circle,
    "nyquist_simulation": dpg.mvPlotMarker_Cross,
    "bode_magnitude_data": dpg.mvPlotMarker_Circle,
    "bode_magnitude_simulation": dpg.mvPlotMarker_Cross,
    "bode_phase_data": dpg.mvPlotMarker_Square,
    "bode_phase_simulation": dpg.mvPlotMarker_Plus,
    "mu_Xps_mu": dpg.mvPlotMarker_Circle,
    "mu_Xps_Xps": dpg.mvPlotMarker_Square,
}


DEFAULT_COLORS: Dict[str, List[float]] = {
    "residuals_real": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "residuals_imaginary": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "nyquist_data": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "nyquist_simulation": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "bode_magnitude_data": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "bode_magnitude_simulation": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "bode_phase_data": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
    "bode_phase_simulation": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "mu_Xps_mu_criterion": [
        255.0,
        255.0,
        255.0,
        128.0,
    ],
    "mu_Xps_mu": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "mu_Xps_mu_highlight": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "mu_Xps_Xps": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "mu_Xps_Xps_highlight": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
}

DEFAULT_TEST_SETTINGS: TestSettings = TestSettings(
    Test.COMPLEX,
    Mode.EXPLORATORY,
    999,
    0.85,
    True,
    True,
    Method.LEASTSQ,
    1000,
)

DEFAULT_FIT_SETTINGS: FitSettings = FitSettings(
    "R([RW]C)",
    Method.AUTO,
    Weight.AUTO,
    1000,
)

DEFAULT_SIMULATION_SETTINGS: SimulationSettings = SimulationSettings(
    "R([RW]C)",
    1.0e-2,
    1.0e5,
    1,
)


VERSION: int = 3


def _parse_v3(dictionary: dict) -> dict:
    # TODO: Update when VERSION is incremented to 4
    return dictionary


def _parse_v2(dictionary: dict) -> dict:
    return _parse_v3(
        {
            "version": 3,
            "auto_backup_interval": dictionary["auto_backup_interval"],
            "num_per_decade_in_simulated_lines": dictionary[
                "num_per_decade_in_simulated_lines"
            ],
            "markers": dictionary["markers"],
            "colors": dictionary["colors"],
            "default_test_settings": dictionary["default_test_settings"],
            "default_fit_settings": dictionary["default_fit_settings"],
            "default_simulation_settings": dictionary["default_simulation_settings"],
            "export_units": dictionary.get("export_units", 1),
            "export_width": dictionary.get("export_width", 10.0),
            "export_height": dictionary.get("export_height", 6.0),
            "export_dpi": dictionary.get("export_dpi", 100),
            "export_preview": dictionary.get("export_preview", 4),
            "export_title": dictionary.get("export_title", True),
            "export_legend": dictionary.get("export_legend", True),
            "export_legend_location": dictionary.get("export_legend_location", 0),
            "export_grid": dictionary.get("export_grid", False),
            "export_tight": dictionary.get("export_tight", False),
            "export_num_per_decade": dictionary.get("export_num_per_decade", 100),
            "export_extension": dictionary.get("export_extension", 4),
            "export_experimental_clear_registry": dictionary.get(
                "export_experimental_clear_registry", True
            ),
            "export_experimental_disable_previews": dictionary.get(
                "export_experimental_disable_previews", False
            ),
        }
    )


def _parse_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    new_color_marker_keys: Dict[str, str] = {
        "real_error": "residuals_real",
        "imag_error": "residuals_imaginary",
        "nyquist_sim": "nyquist_simulation",
        "bode_magnitude_sim": "bode_magnitude_simulation",
        "bode_phase_sim": "bode_phase_simulation",
        "exploratory_mu_criterion": "mu_Xps_mu_criterion",
        "exploratory_mu": "mu_Xps_mu",
        "exploratory_mu_highlight": "mu_Xps_mu_highlight",
        "exploratory_xps": "mu_Xps_Xps",
        "exploratory_xps_highlight": "mu_Xps_Xps_highlight",
    }
    old: str
    new: str
    colors: dict = dictionary.get("colors", {})
    markers: dict = dictionary.get("markers", {})
    for old, new in new_color_marker_keys.items():
        if old in colors:
            colors[new] = colors[old]
            del colors[old]
        if old in markers:
            markers[new] = markers[old]
            del markers[old]
    return _parse_v2(
        {
            "version": 2,
            "auto_backup_interval": dictionary.get("auto_backup_interval", 10),
            "num_per_decade_in_simulated_lines": dictionary[
                "num_per_decade_in_simulated_lines"
            ],
            "markers": dictionary["markers"],
            "colors": dictionary["colors"],
            "default_test_settings": dictionary["default_test_settings"],
            "default_fit_settings": dictionary["default_fit_settings"],
            "default_simulation_settings": dictionary["default_simulation_settings"],
        }
    )


class Config:
    def __init__(self):
        self.config_dir_path: str = join(xdg_config_home(), "DearEIS")
        self.config_path: str = join(self.config_dir_path, "config.json")
        self.auto_backup_interval: int = None  # type: ignore
        self.num_per_decade_in_simulated_lines: int = None  # type: ignore
        self.default_test_settings: TestSettings = None  # type: ignore
        self.default_fit_settings: FitSettings = None  # type: ignore
        self.default_simulation_settings: SimulationSettings = None  # type: ignore
        self.markers: Dict[str, int] = None  # type: ignore
        self.colors: Dict[str, List[float]] = None  # type: ignore
        self.keybindings: List[Keybinding] = None  # type: ignore
        self.export_units: int = 1
        self.export_width: float = 10.0
        self.export_height: float = 6.0
        self.export_dpi: float = 100
        self.export_preview: int = 4
        self.export_title: bool = True
        self.export_legend: bool = True
        self.export_legend_location: int = 0
        self.export_grid: bool = False
        self.export_tight: bool = False
        self.export_num_per_decade: int = 100
        self.export_extension: int = 4
        self.export_experimental_clear_registry: bool = True
        self.export_experimental_disable_previews: bool = False
        self.from_dict(self.default_settings())
        if not exists(self.config_path):
            self.save()
        else:
            try:
                self.load()
            except AssertionError:
                signals.emit(
                    Signal.SHOW_ERROR_MESSAGE,
                    traceback=format_exc(),
                    message="Encountered malformed config! Using defaults...",
                )
            except Exception as e:
                signals.emit(
                    Signal.SHOW_ERROR_MESSAGE,
                    traceback=format_exc(),
                    message="Encountered issue while parsing config! Using defaults...",
                )

    def default_settings(self) -> dict:
        return {
            "version": VERSION,
            "auto_backup_interval": 10,
            "num_per_decade_in_simulated_lines": 100,
            "default_test_settings": DEFAULT_TEST_SETTINGS.to_dict(),
            "default_fit_settings": DEFAULT_FIT_SETTINGS.to_dict(),
            "default_simulation_settings": DEFAULT_SIMULATION_SETTINGS.to_dict(),
            "colors": DEFAULT_COLORS,
            "markers": DEFAULT_MARKERS,
            "keybindings": list(map(lambda _: _.to_dict(), DEFAULT_KEYBINDINGS)),
            "export_units": 1,
            "export_width": 10.0,
            "export_height": 6.0,
            "export_dpi": 100,
            "export_preview": 4,
            "export_title": True,
            "export_legend": True,
            "export_legend_location": 0,
            "export_grid": False,
            "export_tight": False,
            "export_num_per_decade": 100,
            "export_extension": 4,
            "export_experimental_clear_registry": True,
            "export_experimental_disable_previews": False,
        }

    def to_dict(self) -> dict:
        return parse_json(
            dump_json(  # This is done to get new instances in memory
                {
                    "version": VERSION,
                    "auto_backup_interval": self.auto_backup_interval,
                    "num_per_decade_in_simulated_lines": self.num_per_decade_in_simulated_lines,
                    "default_test_settings": self.default_test_settings.to_dict(),
                    "default_fit_settings": self.default_fit_settings.to_dict(),
                    "default_simulation_settings": self.default_simulation_settings.to_dict(),
                    "colors": self.colors,
                    "markers": self.markers,
                    "keybindings": list(map(lambda _: _.to_dict(), self.keybindings)),
                    "export_units": self.export_units,
                    "export_width": self.export_width,
                    "export_height": self.export_height,
                    "export_dpi": self.export_dpi,
                    "export_preview": self.export_preview,
                    "export_title": self.export_title,
                    "export_legend": self.export_legend,
                    "export_legend_location": self.export_legend_location,
                    "export_grid": self.export_grid,
                    "export_tight": self.export_tight,
                    "export_num_per_decade": self.export_num_per_decade,
                    "export_extension": self.export_extension,
                    "export_experimental_clear_registry": self.export_experimental_clear_registry,
                    "export_experimental_disable_previews": self.export_experimental_disable_previews,
                }
            )
        )

    def save(self):
        if not isdir(self.config_dir_path):
            makedirs(self.config_dir_path)
        old_config: str = ""
        fp: IO
        if exists(self.config_path):
            with open(self.config_path, "r") as fp:
                old_config = fp.read().strip()
        new_config: str = dump_json(
            self.to_dict(),
            sort_keys=True,
            indent=2,
        ).strip()
        if new_config == old_config:
            return
        with open(self.config_path, "w") as fp:
            fp.write(new_config)

    def from_dict(self, settings: dict):
        self.auto_backup_interval = settings["auto_backup_interval"]
        self.num_per_decade_in_simulated_lines = settings[
            "num_per_decade_in_simulated_lines"
        ]
        self.default_test_settings = TestSettings.from_dict(
            settings["default_test_settings"]
        )
        self.default_fit_settings = FitSettings.from_dict(
            settings["default_fit_settings"]
        )
        self.default_simulation_settings = SimulationSettings.from_dict(
            settings["default_simulation_settings"]
        )
        self.markers = settings["markers"]
        marker_themes: Dict[str, int] = {
            "residuals_real": themes.residuals.real,
            "residuals_imaginary": themes.residuals.imaginary,
            "nyquist_data": themes.nyquist.data,
            "nyquist_simulation": themes.nyquist.simulation,
            "bode_magnitude_data": themes.bode.magnitude_data,
            "bode_magnitude_simulation": themes.bode.magnitude_simulation,
            "bode_phase_data": themes.bode.phase_data,
            "bode_phase_simulation": themes.bode.phase_simulation,
            "mu_Xps_mu": themes.mu_Xps.mu,
            "mu_Xps_Xps": themes.mu_Xps.Xps,
        }
        key: str
        theme: int
        for key, theme in marker_themes.items():
            themes.update_plot_series_theme_marker(theme, self.markers[key])
        self.colors = settings["colors"]
        color_themes: Dict[str, int] = {
            "residuals_real": themes.residuals.real,
            "residuals_imaginary": themes.residuals.imaginary,
            "nyquist_data": themes.nyquist.data,
            "nyquist_simulation": themes.nyquist.simulation,
            "bode_magnitude_data": themes.bode.magnitude_data,
            "bode_magnitude_simulation": themes.bode.magnitude_simulation,
            "bode_phase_data": themes.bode.phase_data,
            "bode_phase_simulation": themes.bode.phase_simulation,
            "mu_Xps_mu_criterion": themes.mu_Xps.mu_criterion,
            "mu_Xps_mu": themes.mu_Xps.mu,
            "mu_Xps_mu_highlight": themes.mu_Xps.mu_highlight,
            "mu_Xps_Xps": themes.mu_Xps.Xps,
            "mu_Xps_Xps_highlight": themes.mu_Xps.Xps_highlight,
        }
        for key, theme in color_themes.items():
            themes.update_plot_series_theme_color(theme, self.colors[key])
        self.export_units = settings["export_units"]
        self.export_width = settings["export_width"]
        self.export_height = settings["export_height"]
        self.export_dpi = settings["export_dpi"]
        self.export_preview = settings["export_preview"]
        self.export_title = settings["export_title"]
        self.export_legend = settings["export_legend"]
        self.export_legend_location = settings["export_legend_location"]
        self.export_grid = settings["export_grid"]
        self.export_tight = settings["export_tight"]
        self.export_num_per_decade = settings["export_num_per_decade"]
        self.export_extension = settings["export_extension"]
        self.export_experimental_clear_registry = settings[
            "export_experimental_clear_registry"
        ]
        self.export_experimental_disable_previews = settings[
            "export_experimental_disable_previews"
        ]
        self.keybindings = list(map(Keybinding.from_dict, settings["keybindings"]))
        self.validate_keybindings(self.keybindings)

    def check_type(self, user: Any, default: Any, key: str):
        assert type(user) == type(default), (user, default, key)
        if type(default) is list:
            user_types: set = set(list(map(type, user)))
            default_types: set = set(list(map(type, default)))
            assert user_types == default_types, (user_types, default_types, key)
        elif type(default) is dict:
            for key in default:
                if key in user:
                    self.check_type(user[key], default[key], key)

    def merge_dicts(self, user: dict, default: dict) -> dict:
        result: dict = {}
        key: str
        for key in default.keys():
            value: Optional[Any] = user.get(key)
            if value is None:
                result[key] = default[key]
                continue
            self.check_type(value, default[key], key)
            if type(value) is dict:
                result[key] = self.merge_dicts(value, default[key])
            else:
                result[key] = value
        return result

    def load(self):
        with open(self.config_path, "r") as fp:
            dictionary: dict = load_json(fp)
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_v1,
            2: _parse_v2,
            3: _parse_v3,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        self.from_dict(self.merge_dicts(parsers[version](dictionary), self.to_dict()))

    def validate_keybindings(self, keybindings: List[Keybinding]):
        stringified: List[str] = list(map(str, keybindings))
        string: str
        for string in set(stringified):
            count: int = stringified.count(string)
            if count == 1:
                continue
            contexts: List[Set[Context]] = list(
                map(
                    lambda _: set(action_contexts[_.action]),
                    filter(lambda _: str(_) == string, keybindings),
                )
            )
            while contexts:
                a = contexts.pop(0)
                for b in contexts:
                    assert (
                        not a.issuperset(b)
                        and not a.issubset(b)
                        and Context.PROGRAM not in a
                        and Context.PROGRAM not in b
                    ), (
                        "The same keybinding has been applied to multiple actions in the same context or in overlapping contexts:\n- "
                        + "\n- ".join(
                            map(repr, filter(lambda _: str(_) == string, keybindings))
                        )
                    )
        actions: List[Action] = list(map(lambda _: _.action, keybindings))
        action: Action
        for action in set(actions):
            assert actions.count(action) == 1, (
                "Multiple keybindings have been applied to the same action:\n- "
                + "\n- ".join(
                    map(repr, filter(lambda _: _.action == action, keybindings))
                )
            )
